/*
	$Id: main.c,v 1.18 2004/05/30 02:16:36 Treize Exp $
	ROX-MMS - a ROX panel applet front-end to XMMS.
	Copyright (c) 2003 Peter A. Geer <geerp@cs.sunyit.edu>
	
	Adapted from DemoApplet by Stephen Watson.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	oundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "config.h"

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <signal.h>
#include <sys/wait.h>
#include <string.h>
#include <stdarg.h>
#include <unistd.h>

/* These next two are for internationalization */
#ifdef HAVE_LOCALE_H
#include <locale.h>
#endif
#ifdef HAVE_LIBINTL_H
#include <libintl.h>
#endif

#include <gtk/gtk.h>
#include <gdk/gdk.h>
#include <gdk-pixbuf/gdk-pixbuf.h>
#include <xmms/xmmsctrl.h>
#include <libxml/tree.h>
#include <libxml/parser.h>
#include <rox/rox.h>
#include <rox/rox_dnd.h>
#include <rox/options.h>
#include <rox/infowin.h>

#include "pixmaps/play_pause.xpm"
#include "pixmaps/stop.xpm"
#include "pixmaps/next.xpm"
#include "pixmaps/prev.xpm"
#include "pixmaps/eject.xpm"

#define DEBUG              1   /* Allow debug output */
#define UPDATE_FREQUENCY 200
/* These are for the label status. */
#define LABEL_TITLE  0
#define LABEL_TIME   1
#define LABEL_ELAPSE 2
#define LABEL_MODULO 2

/* It's probably a good idea to have constants for the buttons. */
#define BTN_PLAY_PAUSE 0
#define BTN_STOP       1
#define BTN_PREV       2
#define BTN_NEXT       3
#define BTN_EJECT      4

#define MAX_WAIT_ITERS 10

/* Global data. */
static gint xmms_session;
GtkWidget *label;  /* This name was poorly chosen. */
GtkWidget *ev_box; /* Event box for said label. */
static GtkWidget *infowin=NULL;
GtkTooltips *tips=NULL;
guint timer_id;
int label_status=0;
GtkWidget *btn_array[5];
GtkWidget *button_vbox, *main_vbox;

/* Options declarations. */
static Option update_freq_opt;
static Option scroll_len_opt;
static Option display_font_opt;
static Option time_fmt_opt;
static Option date_fmt_opt;
static Option cdmount_opt;
static Option xmms_eject_opt;
static Option eject_cmd_opt;
static Option clock_mode_opt;
static Option buttons_per_row;
static Option no_eject;

/* Declare functions in advance */
static gint button_press(GtkWidget *window, GdkEventButton *bev,
			 gpointer win); /* button press on window */
static void setup_options(void);
static void opts_changed(void);
static void show_info_win(void);   /* Show information box */ 

/* Utility Functions: */
GtkWidget *add_button(GtkWidget *to_hbox, char *icon[], gint btn_code, const char tip_text[]); 
void add_all_buttons(void);
GtkWidget *create_pixmap_d(GtkWidget * widget, char *from[]);
gchar *scroll_string_offset(gchar *title, int offset);
void scroll_string(gchar *title); 
char *get_time_string(char *format);
void set_label_text(GtkWidget *lbl, gchar* text);
void run_xmms(gint session);
gchar *get_xmms_time(gint session, gint pos);

/* Here are my call-back function declarations. */
gint clicked(GtkWidget *widget, gpointer data);
gint timer_callback(void);
static rox_dnd_handle_uris dnd_callback(GtkWidget *widget, GSList *uris,
		gpointer data,	gpointer udata);
static void child_died(int signum);

/* Main.  Here we set up the gui and enter the main loop */
int main(int argc, char *argv[])
{
  GtkWidget *win=NULL;
  GtkWidget *vbox;
  struct sigaction act;
  gchar *app_dir;
  guint32 xid;
#ifdef HAVE_BINDTEXTDOMAIN
  gchar *localedir;
#endif

  /* First things first, set the locale information for time, so that
     strftime will give us a sensible date format... */
#ifdef HAVE_SETLOCALE
  setlocale(LC_TIME, "");
  setlocale (LC_ALL, "");
#endif
  /* What is the directory where our resources are? (set by AppletRun) */
  app_dir = g_strdup(g_getenv("APP_DIR"));
#ifdef HAVE_BINDTEXTDOMAIN
  /* More (untested) i18n support */
  localedir=g_strconcat(app_dir, "/Messages", NULL);
  bindtextdomain(PROJECT, localedir);
  textdomain(PROJECT);
  g_free(localedir);
#endif
  /* Set configuration defaults. */
  xmms_session = 0;
  label_status = 0;  

  /* Initialise X/GDK/GTK */
  gtk_init(&argc, &argv);
  gdk_rgb_init();
  gtk_widget_push_visual(gdk_rgb_get_visual());
  gtk_widget_push_colormap(gdk_rgb_get_cmap());
  
  /* Init choices and read them in */
  choices_init();
  options_init(PROJECT);
  setup_options();

  /* If we have an argument, try to interperet it as an XID */
  if (argc>1)
    xid=atol(argv[1]);
  else
    xid=0;

  if(xid) {
    /* We are an applet, create a plug to go in the filer's socket */
    GtkWidget *plug;

    plug=gtk_plug_new(xid);
    if(!plug) {
      fprintf(stderr, _("%s: failed to plug into socket %s, not a XID?\n"),
	      argv[0], argv[1]);
      exit(1);
    }
    gtk_signal_connect(GTK_OBJECT(plug), "destroy", 
		       GTK_SIGNAL_FUNC(gtk_main_quit), 
		       "WM destroy");
    gtk_signal_connect(GTK_OBJECT(plug), "button_press_event",
		       GTK_SIGNAL_FUNC(button_press), plug);
    gtk_widget_add_events(plug, GDK_BUTTON_PRESS_MASK);
    /* Isn't object oriented code wonderful.  Now that we have done the
       plug specific code, its just a containing widget.  For the rest of the
       program we just pretend it was a window and it all works
    */
    win=plug;
  } else { 
    /* Create window */
    win=gtk_window_new(GTK_WINDOW_TOPLEVEL);
    gtk_widget_set_name(win, PROJECT);
    gtk_window_set_title(GTK_WINDOW(win), _(PROJECT));
    gtk_signal_connect(GTK_OBJECT(win), "destroy", 
		       GTK_SIGNAL_FUNC(gtk_main_quit), 
		       "WM destroy");
    gtk_signal_connect(GTK_OBJECT(win), "delete_event", 
		     GTK_SIGNAL_FUNC(gtk_main_quit), win);
  }

  /* win now refers to either a window or the applet, and it doesn't matter
     which... */
  /* We want to pop up a menu on a button press */
  gtk_signal_connect(GTK_OBJECT(win), "button_press_event", GTK_SIGNAL_FUNC(button_press), win);
		     
  gtk_widget_add_events(win, GDK_BUTTON_PRESS_MASK);

  act.sa_handler = child_died;
  sigemptyset(&act.sa_mask);
  act.sa_flags = SA_NOCLDSTOP;
  sigaction(SIGCHLD, &act, NULL);

  /* A vbox contains widgets arranged vertically */
  vbox=gtk_vbox_new(FALSE, 0);
  gtk_container_add(GTK_CONTAINER(win), vbox);
  gtk_widget_show(vbox);

  ev_box = gtk_event_box_new();
  tips = gtk_tooltips_new();
  rox_dnd_register_uris(GTK_WIDGET(ev_box), 0, 
	                      (rox_dnd_handle_uris)dnd_callback, NULL);
  label=gtk_label_new(_("ROX-MMS"));
  gtk_label_set_justify(GTK_LABEL(label), GTK_JUSTIFY_CENTER);
  gtk_container_add(GTK_CONTAINER(ev_box), label);
  gtk_widget_show(label);
  gtk_box_pack_start(GTK_BOX(vbox), ev_box, TRUE, TRUE, 0);
  gtk_widget_set_events(ev_box, GDK_BUTTON_PRESS_MASK);
  gtk_signal_connect(GTK_OBJECT(ev_box), "button_press_event", GTK_SIGNAL_FUNC(button_press), win);
  gtk_signal_connect(GTK_OBJECT(win), "button_press_event", GTK_SIGNAL_FUNC(button_press), win);
  gtk_widget_show(ev_box);
  gtk_widget_realize(ev_box);
  
  /* Add an vbox for our buttons. */
  button_vbox = gtk_vbox_new(TRUE, 0);
  gtk_box_pack_start(GTK_BOX(vbox), button_vbox, TRUE, TRUE, 0);
  gtk_widget_show(button_vbox); 
  /* Add a pointer to remember the parent of button_vbox. */
  main_vbox = vbox;

  /* Now it's time to add our buttons. */
  add_all_buttons();

  timer_id = gtk_timeout_add((guint32)update_freq_opt.int_value, 
                             (GtkFunction)timer_callback, NULL);

  /* Show our new window/applet */
  gtk_widget_show(win);

  /* Main processing loop */
  gtk_main();

  return 0;
}

/* Keep child processes from staying around as zombies */
static void child_died(int signum)
{
	pid_t	child;
	int	status;
	/* Find out which children exited and allow them to die */
	while (1)
	{
		child = waitpid(-1, &status, WNOHANG);
		if (child == 0 || child == -1)
			break;
	}
}

/* Ad hoc function to add an XMMS control button to the window. */
GtkWidget *add_button(GtkWidget *to_hbox, char *icon[], gint btn_code, 
                const char tip_text[]) {
  
  GtkWidget *button;
  GtkWidget *control_pixmap;
  GtkTooltips *tip;
  
  button = gtk_button_new();
  tip = gtk_tooltips_new();
  gtk_tooltips_set_tip(GTK_TOOLTIPS(tip), button, tip_text, NULL);
  gtk_box_pack_start(GTK_BOX(to_hbox), button, TRUE, TRUE, 0);
  control_pixmap = create_pixmap_d(button, icon);
  gtk_widget_show(control_pixmap);
  gtk_container_add(GTK_CONTAINER(button), control_pixmap);
  gtk_container_set_border_width(GTK_CONTAINER(button), 0);
  gtk_widget_show(button);
  gtk_signal_connect(GTK_OBJECT(button), "clicked", GTK_SIGNAL_FUNC(clicked), GINT_TO_POINTER(btn_code) );
  return button;
  
}

static GtkWidget *hbox_new(void) {
  GtkWidget *ret;	
  ret=gtk_hbox_new(TRUE, 0); 
  gtk_box_pack_start(GTK_BOX(button_vbox), ret, TRUE, TRUE, 0);
  gtk_widget_show(ret); 
  return ret;
}

void add_all_buttons(void) {
	GtkWidget *hbox;
	int added_count = 0;
	hbox = hbox_new();
	btn_array[added_count++] = add_button(hbox, play_pause_xpm, BTN_PLAY_PAUSE, "Play/Pause");
	if (added_count >= buttons_per_row.int_value) hbox = hbox_new();
	btn_array[added_count++] = add_button(hbox, stop_xpm, BTN_STOP, "Stop");
	if (added_count >= buttons_per_row.int_value) hbox = hbox_new();
	btn_array[added_count++] = add_button(hbox, prev_xpm, BTN_PREV, "Previous track");
	if (added_count >= buttons_per_row.int_value) hbox = hbox_new();
	btn_array[added_count++] = add_button(hbox, next_xpm, BTN_NEXT, "Next track");
	if (added_count >= buttons_per_row.int_value && !no_eject.int_value) hbox = hbox_new();
	if (!no_eject.int_value) btn_array[added_count] = add_button(hbox, eject_xpm, BTN_EJECT, "Eject");
}

static void setup_options(void) {
	option_add_int(&update_freq_opt, "update_frequency", 200);
	option_add_int(&scroll_len_opt, "title_length", 15);
	option_add_int(&xmms_eject_opt, "use_xmms_eject", 1);
	option_add_int(&clock_mode_opt, "clock_mode", 0);
	option_add_int(&buttons_per_row, "buttons_per_row", 5);
	option_add_int(&no_eject, "suppress_eject", 0);
	option_add_string(&display_font_opt, "display_font", "Luxi Mono 12");
	option_add_string(&cdmount_opt, "cdrom_mount_path", "/mnt/cdrom");
	option_add_string(&time_fmt_opt, "time_format", "%l:%M %p %Z");
	option_add_string(&date_fmt_opt, "date_format", "%a, %d-%b-%Y");
	option_add_string(&eject_cmd_opt, "eject_command", "eject /dev/cdrom");
	option_add_notify(opts_changed);
}

static void opts_changed(void) {
	if (update_freq_opt.has_changed) {
		gtk_timeout_remove(timer_id);
		timer_id = gtk_timeout_add((guint32)update_freq_opt.int_value, 
		                           (GtkFunction)timer_callback, NULL);
	} 
	if (no_eject.has_changed || buttons_per_row.has_changed) {
		gtk_widget_destroy(button_vbox);
		button_vbox = gtk_vbox_new(TRUE, 0);
		gtk_box_pack_start(GTK_BOX(main_vbox), button_vbox, TRUE, TRUE, 0);
		gtk_widget_show(button_vbox); 
		add_all_buttons();
	}
}

/* Check if XMMS is running.  If it's not, try to start it. */
void run_xmms (gint_session) {
	pid_t child_proc;
	if (!xmms_remote_is_running(xmms_session)) {
		child_proc = fork();
		if (child_proc == 0) {
			execlp("xmms", "xmms", NULL);
			exit(127);
		} 
		while (! xmms_remote_is_running(xmms_session))
			while (g_main_iteration(FALSE));
	}
}

static void play_audio_cd() {
	run_xmms(xmms_session);
	xmms_remote_playlist_clear(xmms_session);
	xmms_remote_playlist_add_url_string(xmms_session, cdmount_opt.value);
	xmms_remote_set_playlist_pos(xmms_session, 0);
	xmms_remote_play(xmms_session);
}

static void quit_xmms(void) {
	xmms_remote_quit(xmms_session);
}

static void start_xmms(void) {
	run_xmms(xmms_session);
}

static void clear_playlist(void) {
	run_xmms(xmms_session);
	xmms_remote_playlist_clear(xmms_session);
}

static void config_win_show(void) {
	options_show();
}

/* Make a destroy-frame into a close, allowing us to re-use the window */
static int trap_frame_destroy(GtkWidget *widget, GdkEvent *event,
			      gpointer data)
{
  /* Change this destroy into a hide */
  gtk_widget_hide(GTK_WIDGET(data));
  return TRUE;
}

/* Show the info window */
static void show_info_win(void)
{
  if(!infowin) {
    infowin=info_win_new(PROJECT, PURPOSE, VERSION, AUTHOR, WEBSITE);
    gtk_signal_connect(GTK_OBJECT(infowin), "delete_event", 
		     GTK_SIGNAL_FUNC(trap_frame_destroy), 
		     infowin);
  }

  gtk_widget_show(infowin);
}

/*
 * Pop-up menu
 */
static GtkItemFactoryEntry menu_items[] = {
	{ _("/Audio CD"), NULL, play_audio_cd, 0, "<StockItem>", GTK_STOCK_CDROM },
	{ _("/Clear Playlist"), NULL, clear_playlist, 0, NULL},
	{ _("/sep"), NULL, NULL, 0, "<Separator>" },
	{ _("/Start XMMS"), NULL, start_xmms, 0, NULL }, 
	{ _("/Quit XMMS"), NULL, quit_xmms, 0, NULL }, 
	{ _("/sep"), NULL, NULL, 0, "<Separator>" },
	{ _("/Options"), NULL, config_win_show, 0, "<StockItem>", GTK_STOCK_PREFERENCES },
	{ _("/Info"), NULL, show_info_win, 0, "<StockItem>", GTK_STOCK_DIALOG_INFO  },
	{ _("/sep"), NULL, NULL, 0, "<Separator>" },
	{ _("/Quit"), NULL, gtk_main_quit, 0, "<StockItem>", GTK_STOCK_QUIT },
};

/* Save user-defined menu accelerators */
#if 0
static void save_menus(void)
{
  char	*menurc;
	
  menurc = choices_find_path_save("menus", PROJECT, TRUE);
  if (menurc) {
    gtk_item_factory_dump_rc(menurc, NULL, TRUE);
    g_free(menurc);
  }
}
#endif

/* Create the pop-up menu */
static GtkWidget *menu_create_menu(GtkWidget *window)
{
	
  GtkWidget *menu;
  GtkItemFactory	*item_factory;
  /*GtkAccelGroup	*accel_group;
  gchar *menurc = NULL;*/
  gint 		n_items = sizeof(menu_items) / sizeof(*menu_items);

  /*accel_group = gtk_accel_group_new();*/

  item_factory = gtk_item_factory_new(GTK_TYPE_MENU, "<system>", NULL);

  gtk_item_factory_create_items(item_factory, n_items, menu_items, NULL);

  /* Attach the new accelerator group to the window. */
  /*gtk_accel_group_attach(GTK_OBJECT(window), accel_group);*/

  menu = gtk_item_factory_get_widget(item_factory, "<system>");

  /* Load any user-defined menu accelerators */
  /*menurc=choices_find_path_load("menus", PROJECT);
  if(menurc) {
    gtk_item_factory_parse_rc(menurc);
    g_free(menurc);
  }*/

  /* Save updated accelerators when we exit */
  /*atexit(save_menus);*/

  return menu;

}

/* Button press on our window/applet */
static gint button_press(GtkWidget *window, GdkEventButton *bev,
			 gpointer win)
{
	static GtkWidget *menu=NULL;
	
	/*  Right-clicking pops up the context menu. */
	if(bev->type==GDK_BUTTON_PRESS && bev->button==3) {
    /* Pop up the menu */
		if(!menu) 
			menu=menu_create_menu(GTK_WIDGET(win));
    	
		gtk_menu_popup(GTK_MENU(menu), NULL, NULL, NULL, NULL,
						bev->button, bev->time);
		return TRUE;
	/* Left-clicking toggles the display. */
	} else if (bev->type == GDK_BUTTON_PRESS && bev->button == 1) {
		label_status = (label_status+1) % LABEL_MODULO;
		return TRUE;
	}
	return FALSE;
}



GtkWidget *create_pixmap_d(GtkWidget * widget, char *from[])
{
	GdkPixbuf *pixbuf;
	GtkWidget *pixmap;
	
	pixbuf = gdk_pixbuf_new_from_xpm_data((const char **) from);
	pixmap = gtk_image_new_from_pixbuf(GDK_PIXBUF(pixbuf));
	
	if (GDK_IS_PIXBUF(pixbuf) ) gdk_pixbuf_unref(pixbuf);

	return pixmap;
}

/* Here we take care of our button presses. */
gint clicked(GtkWidget *widget, gpointer data) {
	int option;
	pid_t child_proc;
	gchar *cmd, *space_pos;
	option = GPOINTER_TO_INT(data);
	if (option != BTN_EJECT) run_xmms(xmms_session);
	switch(option) {
		case BTN_PLAY_PAUSE:
			xmms_remote_play_pause(xmms_session);
			break;
		case BTN_STOP:
			xmms_remote_stop(xmms_session);
			break;
		case BTN_PREV:
			xmms_remote_playlist_prev(xmms_session);
			break;
		case BTN_NEXT:
			xmms_remote_playlist_next(xmms_session);
			break;
		case BTN_EJECT:
			if (xmms_eject_opt.int_value) {
				run_xmms(xmms_session);
				xmms_remote_eject(xmms_session);
			} else {
				child_proc = fork();
				if (child_proc == 0) {
					space_pos = g_strstr_len(eject_cmd_opt.value, 128, _(" "));
					space_pos[0] = '\0';
					cmd = g_strdup(eject_cmd_opt.value);
					space_pos[0] = _(' ');
					execlp(cmd, eject_cmd_opt.value, NULL);
					exit(127);
				} 		
			}
			break;
	}
	return 0;
}

gint timer_callback(void) {

	/* For displaying the date and time. */
	char* time_string = NULL;
	char *date_string = NULL;

	gint pos = -1;
	gint list_len;
	static gchar *title_string;
	gchar *current_title=NULL;
	static gchar *raw_title=NULL;
	gchar *track_time_string = NULL;
	gchar *temp;
	
	/* In case XMMS is running, we need to display the track info... */
	if(xmms_remote_is_running(xmms_session)) {
	
		/* Get the playlist length and return if there's no track. */
		run_xmms(xmms_session);
		list_len = xmms_remote_get_playlist_length(xmms_session);
		if (list_len <= 0) {
			set_label_text(label, _("No track..."));
			return TRUE;
		}

		/* Get somw strings that we'll use in both cases. */
		pos = xmms_remote_get_playlist_pos(xmms_session);
		track_time_string = get_xmms_time(xmms_session, pos);
		current_title = xmms_remote_get_playlist_title(xmms_session, pos);
		
		switch (label_status) {
			case LABEL_TIME:
				/* If not in clock mode, display the track time.*/
				if (! clock_mode_opt.int_value) {
					set_label_text(label, track_time_string);
					break;
				}
				/* If we *are* in clock mode, we display the track title
				   here, instead of in title mode.
					NOTE: This case uses drop-through evaluation.  YHBW. */
				
			case LABEL_TITLE:
				/* Check if the title string has changed. */
				run_xmms(xmms_session);
				if (raw_title != NULL && 
				    g_ascii_strcasecmp(current_title, raw_title) == 0)
				{
					/* If not, scroll the title. */
					if (strlen(title_string) > scroll_len_opt.int_value)
						scroll_string(title_string);
				} else {
					/* If so, free old strings and set new one, adding 
					   seperator if needed. */
					if (raw_title) g_free(raw_title);
					if (title_string) g_free(title_string);
					raw_title = g_strdup(current_title);
					if (strlen(raw_title) > scroll_len_opt.int_value) {		
						title_string = g_strconcat(_(" * "), raw_title, NULL);
					} else {
						title_string = g_strdup(raw_title);
					}
				}
				/* If we're in clock mode and in LABEL_TITLE state, then set
				   the display to the time instead of the title. */
				if (clock_mode_opt.int_value && label_status == LABEL_TITLE) {
					time_string = get_time_string(time_fmt_opt.value);
					set_label_text(label, time_string);
					g_free(time_string);
				} else {
					temp = g_strndup(title_string, scroll_len_opt.int_value);
					set_label_text(label, temp);
					g_free(temp);
				}
				break;

			default:
				set_label_text(label, _("No track..."));
				break;
		} 

		temp = g_strconcat(current_title, " - ", track_time_string, NULL);
		gtk_tooltips_set_tip(tips, ev_box, temp, NULL);
		/* Free all those strings we allocated. */
		g_free(temp);
		g_free(track_time_string);
		g_free(current_title);
	
	/* If XMMS is not running, we need to display the time. */
	} else {
		
		time_string = get_time_string(time_fmt_opt.value);
		date_string = get_time_string(date_fmt_opt.value);
		
		switch (label_status) {
			case LABEL_TITLE:
				set_label_text(label, time_string);
				gtk_tooltips_set_tip(tips, ev_box, date_string, NULL);
				break;

			case LABEL_TIME:
				set_label_text(label, date_string);
				gtk_tooltips_set_tip(tips, ev_box, time_string, NULL);
				break;

			default:
				set_label_text(label, time_string);
				gtk_tooltips_set_tip(tips, ev_box, date_string, NULL);
				break;
		}

		g_free(time_string);
		g_free(date_string);
		gtk_tooltips_enable(tips);

	}

	return TRUE;
}

/* This function will "scroll" the string by offset characters. */
gchar *scroll_string_offset(gchar *title, int offset) {

	gchar *new_end = NULL;
	gchar *new_start = NULL;
	gchar *new_string = NULL;

	/* Nothing to do for zero offset or a null string. */
	if (offset == 0 || title == NULL) return title;

	/* Get the part that will be moved to the end. */
	new_end = g_strndup(title, offset);
	/* Now get the part that will be moved to the front. */
	new_start = g_strdup((gchar*)title+offset);
	/* Now we can put them together backwards. */
	new_string = g_strconcat(new_start, new_end, NULL);

	g_free(new_start);
	g_free(new_end);
	
	return new_string;

}

/* Simpler version that scrolls by one character. */
void scroll_string(gchar *title) {

	gchar first;
	int len, i;

	first = title[0];
	len = strlen(title);
	/*memmove(title, (char*)title+1, len-1);*/
	for (i=1; i<len; i++) 
		title[i-1] = title[i];
	title[len-1] = first;

}

/* Get the elapsed/total time string for the current track. */
gchar *get_xmms_time(gint session, gint pos) {
	
	gint time, t_min, t_sec, track_length, l_min, l_sec;
	gchar *ret = NULL;
	
	run_xmms(xmms_session);
	time = xmms_remote_get_output_time(session);
	run_xmms(xmms_session);
	track_length = xmms_remote_get_playlist_time(session, pos);
	
	/* The time appears to be in milliseconds, so we'll convert to straight seconds. */
	t_min = (time - (time % 60000)) / 60000;
	t_sec = (time % 60000) / 1000;
	l_min = (track_length - (track_length % 60000)) / 60000;
	l_sec = (track_length % 60000) / 1000;
	ret = g_strdup_printf("%d:%02d / %d:%02d", t_min, t_sec, l_min, l_sec);

	return ret;
}

/* Return a string for the current time. */
char *get_time_string(char *format) {
	time_t ct;
	struct tm *ctm;
	char *time_string=NULL;
	int size;
	/* Can this possibly be less than the actual size? */
	size = strlen(format) * 3;
	time_string=calloc(size, sizeof(char));
	time(&ct);
	ctm = localtime(&ct);
	strftime(time_string, size, format, ctm); /*"%l:%M %p %Z", ctm);*/
	if (time_string == NULL) return NULL;
	free(ctm);
	return time_string;
}

static rox_dnd_handle_uris dnd_callback(GtkWidget *widget, GSList *uris,
		gpointer data,	gpointer udata) {	
	run_xmms(xmms_session);
	xmms_remote_playlist_add(xmms_session, (GList*)uris);
}

void set_label_text(GtkWidget *lbl, gchar* text) {
	
	int len = strlen(text);
	int diff = 0;
	int i = 0;
	gchar *tmp = NULL;
	gchar* lbl_txt = NULL;
	
	diff = scroll_len_opt.int_value - len;

	if (diff > 0) {
		tmp = g_strnfill(scroll_len_opt.int_value, ' ');
		strcpy(tmp+(diff/2), text);
		for (i=0; i<scroll_len_opt.int_value; i++)
			if (tmp[i]=='\0') tmp[i]=' ';
	} else tmp = text;
	
	lbl_txt = g_strconcat("<span font_desc='", display_font_opt.value,
	                      "'>", tmp, "</span>", NULL);
/*printf("%s : %s\n", display_font_opt.value, lbl_txt);*/
	if (diff > 0) g_free(tmp);
	gtk_label_set_markup(GTK_LABEL(lbl), lbl_txt);
	g_free(lbl_txt);
	
}
